"""
This is a complete rewrite of the python ftplib FTP client with FTPes and FTPs support on any TCP port.

© 2020 - MCO System - https://www.mcos.nc
"""
import re
import socket
import logging
from socket import _GLOBAL_DEFAULT_TIMEOUT
from laboro_ftp.clients.errors import Error, ErrorPerm, ErrorProto, ErrorReply, ErrorTemp

CRLF = '\r\n'
B_CRLF = b'\r\n'
MSG_OOB = 0x1


def sanitize(message):
  if message[:5] in {'pass ', 'PASS '}:
    i = len(message.rstrip('\r\n'))
    message = message[:5] + '*' * (i - 5) + message[i:]
  return repr(message)


def parse150(resp):
  if resp[:3] != '150':
    raise ErrorReply(resp)
  reg_150 = re.compile(r"150 .* \((\d+) bytes\)", re.IGNORECASE | re.ASCII)
  m = reg_150.match(resp)
  if not m:
    return None
  return int(m.group(1))


def parse227(resp):
  if resp[:3] != '227':
    raise ErrorReply(resp)
  reg_227 = re.compile(r'(\d+),(\d+),(\d+),(\d+),(\d+),(\d+)', re.ASCII)
  m = reg_227.search(resp)
  if not m:
    raise ErrorProto(resp)
  numbers = m.groups()
  host = '.'.join(numbers[:4])
  port = (int(numbers[4]) << 8) + int(numbers[5])
  return host, port


def parse229(resp, peer):
  if resp[:3] != '229':
    raise ErrorReply(resp)
  left = resp.find('(')
  if left < 0:
    raise ErrorProto(resp)
  right = resp.find(')', left + 1)
  if right < 0:
    raise ErrorProto(resp)
  if resp[left + 1] != resp[right - 1]:
    raise ErrorProto(resp)
  parts = resp[left + 1:right].split(resp[left + 1])
  if len(parts) != 5:
    raise ErrorProto(resp)
  host = peer[0]
  port = int(parts[3])
  return host, port


def parse257(resp):
  if resp[:3] != '257':
    raise ErrorReply(resp)
  if resp[3:5] != ' "':
    return ''
  dirname = ''
  i = 5
  n = len(resp)
  while i < n:
    c = resp[i]
    i = i + 1
    if c == '"':
      if i >= n or resp[i] != '"':
        break
      i = i + 1
    dirname = dirname + c
  return dirname


def print_line(line):
  print(line)


class FtpSimple:
  """
  Basic FTP client without SSL support. This class is the base class for Ftps and Ftpes classes.
  """

  ENCODING = 'latin-1'
  maxline = 8192
  sock = None
  af = None
  file = None
  welcome = None
  passive = 1

  def __init__(self, hostname, port=21,
               username='anonymous', password='anonymous',
               timeout=_GLOBAL_DEFAULT_TIMEOUT):
    self.hostname = hostname
    self.port = port
    self.username = username
    self.password = password
    self.timeout = timeout

  def connect(self):
    logging.debug("Connecting...")
    if self.hostname is not None:
      self.sock = socket.create_connection((self.hostname, self.port),
                                           self.timeout)
      self.af = self.sock.family
      self.file = self.sock.makefile('r', encoding=self.ENCODING)
      logging.debug(f"[WELCOME] {self._getresp()}")
      logging.debug("[=] connected")
      return True
    raise Error("Remote host not defined")

  def login(self):
    resp = self.sendcmd(f"USER {self.username}")
    if resp[0] == '3':
      resp = self.sendcmd(f"PASS {self.password}")
    if resp[0] != '2':
      raise ErrorReply(resp)
    logging.debug("[=] Logged in")
    return resp

  def _putline(self, line):
    if '\r' in line or '\n' in line:
      raise ValueError("Illegal newline character should not be contained")
    line = line + CRLF
    self.sock.sendall(line.encode(self.ENCODING))

  def _getline(self):
    line = self.file.readline(self.maxline + 1)
    if len(line) > self.maxline:
      raise Error(f"got more than {self.maxline} bytes")
    if not line:
      raise EOFError
    if line[-2:] == CRLF:
      line = line[:-2]
    elif line[-1:] in CRLF:
      line = line[:-1]
    return line

  def _putcmd(self, line):
    logging.debug(f"[CMD] {sanitize(line)}")
    self._putline(line)

  def _getmultiline(self):
    line = self._getline()
    if line[3:4] == '-':
      code = line[:3]
      while 1:
        nextline = self._getline()
        line += (f"\n{nextline}")
        if nextline[:3] == code and nextline[3:4] != '-':
          break
    return line

  def _getresp(self):
    resp = self._getmultiline()
    logging.debug(f"[RESP] {sanitize(resp)}")
    c = resp[:1]
    if c in {'1', '2', '3'}:
      return resp
    if c == '4':
      raise ErrorTemp(resp)
    if c == '5':
      raise ErrorPerm(resp)
    raise ErrorProto(resp)

  def _voidresp(self):
    resp = self._getresp()
    if resp[:1] != '2':
      raise ErrorReply(resp)
    return resp

  def sendcmd(self, cmd):
    self._putcmd(cmd)
    return self._getresp()

  def voidcmd(self, cmd):
    self._putcmd(cmd)
    return self._voidresp()

  def set_pasv(self, value):
    self.passive = value

  def sendport(self, host, port):
    hbytes = host.split('.')
    pbytes = [repr(port // 256), repr(port % 256)]
    tbytes = hbytes + pbytes
    cmd = f"PORT {','.join(tbytes)}"
    return self.voidcmd(cmd)

  def sendeprt(self, host, port):
    '''Send an EPRT command with the current host and the given port number.'''
    af = 0
    if self.af == socket.AF_INET:
      af = 1
    if self.af == socket.AF_INET6:
      af = 2
    if af == 0:
      raise ErrorProto('unsupported address family')
    fields = ['', repr(af), host, repr(port), '']
    cmd = f"EPRT {'|'.join(fields)}"
    return self.voidcmd(cmd)

  def makeport(self):
    err = None
    sock = None
    for res in socket.getaddrinfo(None, 0,
                                  self.af,
                                  socket.SOCK_STREAM, 0,
                                  socket.AI_PASSIVE):
      af, socktype, proto, canonname, sa = res
      try:
        sock = socket.socket(af, socktype, proto)
        sock.bind(sa)
      except OSError as _:
        err = _
        if sock:
          sock.close()
        sock = None
        continue
      break
    if sock is None:
      if err is not None:
        raise err
      raise OSError("getaddrinfo returns an empty list")
    sock.listen(1)
    port = sock.getsockname()[1]
    host = self.sock.getsockname()[0]
    if self.af == socket.AF_INET:
      resp = self.sendport(host, port)
    else:
      resp = self.sendeprt(host, port)
      sock.settimeout(self.timeout)
    return sock

  def makepasv(self):
    if self.af == socket.AF_INET:
      host, port = parse227(self.sendcmd('PASV'))
    else:
      host, port = parse229(self.sendcmd('EPSV'), self.sock.getpeername())
    return host, port

  def ntransfercmd(self, cmd, rest=None):
    size = None
    if self.passive:
      host, port = self.makepasv()
      logging.debug(f"[PASV] {host}:{port}")
      conn = socket.create_connection((host, port), self.timeout)
      try:
        if rest is not None:
          self.sendcmd(f"REST {rest}")
        resp = self.sendcmd(cmd)
        if resp[0] == '2':
          resp = self._getresp()
        if resp[0] != '1':
          raise ErrorReply(resp)
      except Exception:
        conn.close()
        raise
    else:
      with self.makeport() as sock:
        if rest is not None:
          self.sendcmd(f"REST {rest}")
        resp = self.sendcmd(cmd)
        if resp[0] == '2':
          resp = self._getresp()
        if resp[0] != '1':
          raise ErrorReply(resp)
        conn, sockaddr = sock.accept()
        if self.timeout is not _GLOBAL_DEFAULT_TIMEOUT:
          conn.settimeout(self.timeout)
    if resp[:3] == '150':
      size = parse150(resp)
    return conn, size

  def transfercmd(self, cmd, rest=None):
    return self.ntransfercmd(cmd, rest)[0]

  def retrbinary(self, cmd, callback, blocksize=8192, rest=None):
    self.voidcmd('TYPE I')
    with self.transfercmd(cmd, rest) as conn:
      while 1:
        data = conn.recv(blocksize)
        if not data:
          break
        callback(data)
    return self._voidresp()

  def retrlines(self, cmd, callback=None):
    if callback is None:
      callback = print_line
    resp = self.sendcmd('TYPE A')
    with self.transfercmd(cmd) as conn, \
         conn.makefile('r', encoding=self.ENCODING) as fp:
      while 1:
        line = fp.readline(self.maxline + 1)
        if len(line) > self.maxline:
          raise Error(f"got more than {self.maxline} bytes")
        logging.debug(f"[RETR] {repr(line)}")
        if not line:
          break
        if line[-2:] == CRLF:
          line = line[:-2]
        elif line[-1:] == '\n':
          line = line[:-1]
        callback(line)
    return self._voidresp()

  def storbinary(self, cmd, fp, blocksize=8192, callback=None, rest=None):
    self.voidcmd('TYPE I')
    with self.transfercmd(cmd, rest) as conn:
      while 1:
        buf = fp.read(blocksize)
        if not buf:
          logging.debug("[STOR] EOF")
          break
        conn.sendall(buf)
        if callback:
          callback(buf)
    return self._voidresp()

  def storlines(self, cmd, fp, callback=None):
    self.voidcmd('TYPE A')
    with self.transfercmd(cmd) as conn:
      while 1:
        buf = fp.readline(self.maxline + 1)
        if len(buf) > self.maxline:
          raise Error(f"got more than {self.maxline} bytes")
        if not buf:
          break
        if buf[-2:] != B_CRLF:
          if buf[-1] in B_CRLF:
            buf = buf[:-1]
            buf = buf + B_CRLF
        conn.sendall(buf)
        if callback:
          callback(buf)
    return self._voidresp()

  def nlst(self, *args):
    cmd = 'NLST'
    for arg in args:
      cmd = f"{cmd} {arg}"
    files = []
    self.retrlines(cmd, files.append)
    return files

  def dir(self, *args):
    cmd = 'LIST'
    func = None
    if args[-1:] and isinstance(args[-1], str):
      args, func = args[:-1], args[-1]
    for arg in args:
      if arg:
        cmd = f"{cmd} {arg}"
    self.retrlines(cmd, func)

  def mlsd(self, path="", facts=None):
    if facts is None:
      facts = []
    if facts:
      self.sendcmd("OPTS MLST " + ";".join(facts) + ";")
    if path:
      cmd = f"MLSD {path}"
    else:
      cmd = "MLSD"
    lines = []
    self.retrlines(cmd, lines.append)
    for line in lines:
      facts_found, _, name = line.rstrip(CRLF).partition(' ')
      entry = {}
      for fact in facts_found[:-1].split(";"):
        key, _, value = fact.partition("=")
        entry[key.lower()] = value
      yield (name, entry)

  def rename(self, fromname, toname):
    resp = self.sendcmd(f"RNFR {fromname}")
    if resp[0] != '3':
      raise ErrorReply(resp)
    return self.voidcmd(f"RNTO {toname}")

  def delete(self, filename):
    resp = self.sendcmd(f"DELE {filename}")
    if resp[:3] in {'250', '200'}:
      return resp
    raise ErrorReply(resp)

  def cwd(self, dirname):
    if dirname == '..':
      try:
        return self.voidcmd('CDUP')
      except ErrorPerm as msg:
        if msg.args[0][:3] != '500':
          raise
    elif dirname == '':
      dirname = '.'
    cmd = f"CWD {dirname}"
    return self.voidcmd(cmd)

  def size(self, filename):
    resp = self.sendcmd(f"SIZE {filename}")
    if resp[:3] == '213':
      s = resp[3:].strip()
      return int(s)
    raise ErrorReply(resp)

  def mkd(self, dirname):
    resp = self.voidcmd(f"MKD {dirname}")
    if not resp.startswith('257'):
      return ''
    return parse257(resp)

  def rmd(self, dirname):
    return self.voidcmd(f"RMD {dirname}")

  def pwd(self):
    resp = self.voidcmd('PWD')
    if not resp.startswith('257'):
      return ''
    return parse257(resp)

  def quit(self):
    resp = self.voidcmd('QUIT')
    self.close()
    return resp

  def close(self):
    logging.debug("[CLOSE]")
    self.file.close()
    self.sock.close()



import os
import re
import logging
from laboro.error import LaboroError
from laboro.module import Module
from laboro_ftp.clients.ftp import FtpSimple
from laboro_ftp.clients.ftps import Ftps
from laboro_ftp.clients.ftpes import Ftpes


class Ftp(Module):
  """This class is derived from the ``laboro.module.Module`` base class.

  Its purpose is to provide a FTP client able to support all type of operations with or without implicit or explicit TLS support.

  Arguments:

    args: A dictionary specifying the FTP connection parameters and client type.

      ``username``: String. Optional, default to 'anonymous'.
      ``password``: String. Optional, default to 'anonymous'.
      ``host``: String. The remote host to connect to.
      ``port``: Int. Optional. The TCP remote port to connect to. Default: ``21``.
      ``tls``: Boolean. Optional, default to True. If set to False, the FTP connection will not be encrypted.
      ``explicit``: Boolean. Optional, When set to True (default), the FTP connection will use explicit TLS (FTPes). If set to False, the connection will use implicit TLS (FTPs).
      ``ssl_verify``: Boolean. Optional. Default to True. If set to False, remote server SSL certificate will not be verified.
      ``ssl_version``: String. Optional. A string specifying the TLS version to use (Usualy one of ``TLSv1_3`` or  ``TLSv1_2``). Default to `ssl.PROTOCOL_TLS_CLIENT` which will trigger negotiation between client and server for highest possible security.
  """

  def __init__(self, context, args=None):
    super().__init__(filepath=__file__, context=context, args=args)
    self.client = None
    self.connected = False

  def __enter__(self):
    super().__enter__()
    if self.args.get("tls"):
      if self.args.get("explicit"):
        self.client = Ftpes(self.args.get("host"),
                            self.args.get("port"),
                            self.args.get("username"),
                            self.args.get("password"),
                            self.args.get("ssl_verify"),
                            self.args.get("ssl_version"))
      else:
        self.client = Ftps(self.args.get("host"),
                           self.args.get("port"),
                           self.args.get("username"),
                           self.args.get("password"),
                           self.args.get("ssl_verify"),
                           self.args.get("ssl_version"))
    else:
      self.client = FtpSimple(self.args.get("host"),
                              self.args.get("port"),
                              self.args.get("username"),
                              self.args.get("password"))
    self._connect()
    return self

  def __exit__(self, type_err, value, traceback):
    self._close()

  def _connect(self):
    """Connect to the remote server using the specified connection information.
    """
    logging.info(f"Connecting to {self.args.get('host')}:{self.args.get('port')}")
    self.connected = self.client.connect()
    if self.connected:
      self.client.login()

  def _close(self):
    """Close the connection to remote server.
    """
    logging.info(f"Closing connection to {self.args.get('host')}:{self.args.get('port')}")
    if self.client is not None:
      self.client.quit()
      self.connected = False

  @Module.laboro_method
  def get_file(self, remote_file, local_dir,
               create_dirs=False, exit_on_error=True):
    """
    Get file from remote server ``remote_file`` and store it to ``local_dir``.
    If ``create_dir`` is set to ``True``, a directory structure matching the remote directory hierarchy will be created locally.

    Arguments:
      ``remote_file``: The remote filepath to get the file from.
      ``local_dir``: The local directory where to store the local copy of the file.
      ``create_dirs``: Boolean. Optional. Create an exact copy of the remote directory structure. Default: ``False``
      ``exit_on_error``: Boolean. Optional. If set to ``False``, any error encountered will only be logged as a warning. Default: ``True``.
    """
    local_fname = os.path.join(local_dir, os.path.basename(remote_file))
    if create_dirs:
      local_fname = os.path.join(local_dir, remote_file)
      self._create_local_dirs(remote_file, local_dir)
    logging.info(f"Getting remote file: {remote_file}")
    try:
      self.client.retrbinary(f"RETR {remote_file}",
                             open(local_fname, 'wb').write)
    except Exception as err:
      err_msg = f"[FtpTransferError] {err.__class__.__name__} {err}"
      if exit_on_error:
        logging.critical(err_msg)
        raise LaboroError(err_msg) from err
      logging.warning(err_msg)

  @Module.laboro_method
  def put_file(self, local_file, remote_dir="/",
               create_dirs=False, exit_on_error=True):
    """
    Send local file ``local_file`` to remote directory ``remote_dir``.
    If ``create_dirs`` is set to ``True``, a directory structure matching the local directory hierarchy will be created on the remote host.

    Arguments:
      ``local_file``: String. Local filename full path.
      ``remote_dir``: String. Optional. Remote directory where to store the file. Default to the remote root directory.
      ``create_dirs``: Boolean. Optional. Create all missing directory components found in ``remote_dir``. Default: ``False``.
      ``exit_on_error``: Boolean. Optional. If set to ``False``, any error encountered will only be logged as a warning. Default: ``True``.
    """
    remote_dir = remote_dir.lstrip("/")
    remote_file = os.path.basename(local_file)
    remote_fname = os.path.join(remote_dir, remote_file)
    if create_dirs:
      self._create_remote_dirs(remote_dir=remote_dir,
                               exit_on_error=exit_on_error)
      remote_fname = os.path.join(remote_dir, remote_file)
    logging.info(f"Sending file to: {remote_fname}")
    try:
      self.client.storbinary(f"STOR {remote_fname}", open(local_file, "rb"))
    except Exception as err:
      err_msg = f"[FtpTransferError] {err.__class__.__name__} {err}"
      if exit_on_error:
        logging.critical(err_msg)
        raise LaboroError(err_msg) from err
      logging.warning(err_msg)

# -----------------------------------------------------------------------------
# Local operations
# -----------------------------------------------------------------------------
  @Module.laboro_method
  def list_local_dir(self, local_dir, file_type="f", pattern=".*"):
    """
    Get the complete list of files or dirs found in the specified local directory and matching the specified file type (f or d) and the specified regular expression pattern.

    Note:
      - This method is **NOT** recursive.
      - The workflow main workspace directory and workflow session workspace directory component of path will be automatically stripped out from

    Arguments:
      ``local_dir``: String. The full path of the local directory to explore. This option is mandatory and has no default value.
      ``file_type``: String. Optional. If set to ``d``, the returned list will only contain directory names. When set to ``f``, the returned list will contain only file names. Default: ``f``.
      pattern: String. A regular expression to filter returned values. Default to ``.*``
    """
    dirs = list()
    files = list()
    reg = re.compile(pattern)
    logging.info(f"Exploring {local_dir}")
    try:
      for l_file in os.listdir(local_dir):
        l_full_fname = os.path.join(local_dir, l_file)
        if os.path.isdir(l_full_fname) and re.match(reg, l_full_fname):
          dirs.append(l_full_fname)
        elif os.path.isfile(l_full_fname) and re.match(reg, l_full_fname):
          files.append(l_full_fname)
      if file_type == "f":
        return files
      elif file_type == "d":
        return dirs
    except Exception as err:
      err_msg = f"[FtpLocalError] {err.__class__.__name__} {err}"
      logging.critical(err_msg)
      raise LaboroError(err_msg) from err

  @Module.laboro_method
  def local_mirror(self, remote_dir, local_dir, exit_on_error=True):
    """Make a local copy of a remote directory recursively.

    Arguments:
      ``remote_dir``: String. The remote directory full path.
      ``local_dir``: String. The local directory full path.
      ``skip``: Int. Number of path subdirectory to skip. Default to ``0``.
      ``exit_on_error``: Boolean. When set to ``False``, transfer errors are ignored. If set to ``True``, raise error at first transfer error. Default to ``True``.
    """
    for r_file in self.list_remote_dir(remote_dir=remote_dir,
                                       file_type="f", pattern=".*"):
      self.get_file(remote_file=r_file,
                    local_dir=local_dir,
                    create_dirs=True,
                    exit_on_error=exit_on_error)
    for r_dir in self.list_remote_dir(remote_dir=remote_dir,
                                      file_type="d", pattern=".*"):
      self.local_mirror(remote_dir=r_dir,
                        local_dir=local_dir,
                        exit_on_error=exit_on_error)

  def delete_local_file(self, local_file, exit_on_error=True):
    """
    Delete specified file from local host.

    Arguments:
      ``local_file``: The local filepath to delete.
      ``exit_on_error``: Boolean. Optional. If set to ``False``, any error encountered will only be logged as a warning. Default: ``True``.
    """
    logging.info(f"Removing local file: {local_file}")
    try:
      os.remove(local_file)
    except Exception as err:
      err_msg = f"[FtpLocalOSError] {err.__class__.__name__}: {err}"
      if exit_on_error:
        raise LaboroError(err_msg) from err
      logging.warning(err_msg)

  def delete_local_dir(self, local_dir, recursive=False, exit_on_error=True):
    """
    Delete specified directory on the local host and its content.
    Whatever the value of the ``recursive``argument, all files within the specified directory will be deleted.
    When ``recursive`` is set to ``True``, all sub directories and their content are deleted.

    Arguments:
      ``local_dir``: The directory to delete.
      ``recursive``: If set to ``True``, delete directory and all its content.
      ``exit_on_error``: Boolean. Optional. If set to ``False``, any error encountered will only be logged as a warning. Default: ``True``.
    """
    logging.info(f"Removing local dir: {local_dir}")

    try:
      if recursive:
        for l_dir in self.list_local_dir(local_dir=local_dir, file_type="d"):
          self.delete_local_dir(local_dir=l_dir,
                                recursive=recursive,
                                exit_on_error=exit_on_error)
      for l_file in self.list_local_dir(local_dir=local_dir):
        self.delete_local_file(local_file=l_file, exit_on_error=exit_on_error)
      os.rmdir(local_dir)
    except Exception as err:
      err_msg = f"[FtpLocalOSError] {err.__class__.__name__}: {err}"
      if exit_on_error:
        raise LaboroError(err_msg) from err
      logging.warning(err_msg)

  def _create_local_dirs(self, remote_file, local_dir):
    """
    Create a copy of the remote directory structure for a remote file path.

    Arguments:
      ``remote_file``: String. A remote file path.
      ``local_dir``: String. The local directory within duplicate the remote directory tree.
    """
    hierarchy = os.path.join(*remote_file.split("/")[:-1])
    hierarchy = os.path.join(local_dir, hierarchy)
    logging.info(f"Creating local dir: {hierarchy}")
    try:
      os.makedirs(hierarchy, exist_ok=True)
    except Exception as err:
      err_msg = f"[FtpOSError] {err.__class__.__name__} {err}"
      logging.error(err_msg)
      raise LaboroError(err_msg) from err

# -----------------------------------------------------------------------------
# Remote operations
# -----------------------------------------------------------------------------
  @Module.laboro_method
  def list_remote_dir(self, remote_dir="/", file_type="f", pattern=".*"):
    """
    Get the complete list of files or dirs found in the specified remote directory and matching the specified file type (f or d) and the specified regular expression pattern.

    Note: This method is **NOT** recursive.

    Arguments:
      ``remote_dir``: String. Optional. The full path of the remote directory to explore. Default to remote root dir.
      ``file_type``: String. Optional. If set to ``d``, the returned list will only contain directory names. When set to ``f``, the returned list will contain only file names. Default: ``f``.
      pattern: String. A regular expression to filter returned values. Default to ``.*``
    """
    content = list()
    logging.info(f"Exploring {remote_dir}")
    try:
      self.client.retrlines(f"LIST {remote_dir}", content.append)
      if file_type == 'f':
        reg = re.compile("^-.*")
      elif file_type == 'd':
        reg = re.compile("^d.*")
      content = [os.path.join(remote_dir, f.split()[-1]) for f in content if re.match(reg, f)]
      reg = re.compile(pattern)
      return [f.split()[-1] for f in content if re.match(reg, f)]
    except Exception as err:
      err_msg = f"[FtpRemoteError] {err.__class__.__name__} {err}"
      logging.critical(err_msg)
      raise LaboroError(err_msg) from err

  @Module.laboro_method
  def remote_mirror(self, local_dir, remote_dir, exit_on_error=True):
    """Make a remote copy of a local directory recursively.

    Arguments:
      ``local_dir``: String. The local directory full path.
      ``remote_dir``: String. The remote directory full path.
      ``exit_on_error``: Boolean. When set to ``False``, transfer errors are ignored. If set to ``True``, raise error at first transfer error. Default to ``True``.
    """
    wksppath_ptn = re.compile(f"^{self.context.workspace.workspace_path}")
    wkspdir_ptn = re.compile(f"^{self.context.workspace.workspace_dir}")
    r_dir = re.sub(wksppath_ptn, "", local_dir)
    r_dir = re.sub(wkspdir_ptn, "", r_dir)
    r_dir = os.path.join(remote_dir, r_dir)
    for l_file in self.list_local_dir(local_dir=local_dir,
                                      file_type="f",
                                      pattern=".*"):
      self.put_file(local_file=l_file,
                    remote_dir=remote_dir,
                    create_dirs=True,
                    exit_on_error=exit_on_error)
    for l_dir in self.list_local_dir(local_dir=local_dir,
                                     file_type="d",
                                     pattern=".*"):
      self.remote_mirror(local_dir=l_dir,
                         remote_dir=os.path.join(r_dir,
                                                 os.path.basename(l_dir)),
                         exit_on_error=exit_on_error)

  @Module.laboro_method
  def delete_remote_file(self, remote_file, exit_on_error=True):
    """
    Delete specified file from remote server.

    Arguments:
      ``remote_file``: The remote filepath to delete.
      ``exit_on_error``: Boolean. Optional. If set to ``False``, any error encountered will only be logged as a warning. Default: ``True``.
    """
    logging.info(f"Deleting remote file: {remote_file}")
    try:
      self.client.delete(remote_file)
    except Exception as err:
      err_msg = f"[FtpTransferError] {err.__class__.__name__} {err}"
      if exit_on_error:
        logging.critical(err_msg)
        raise LaboroError(err_msg) from err
      logging.warning(err_msg)

  @Module.laboro_method
  def delete_remote_dir(self, remote_dir, recursive=False, exit_on_error=True):
    """
    Delete specified  directory from remote server.

    Arguments:
      ``remote_dir``: The remote directory to delete.
      ``recursive``: If set to ``True``, delete directory and all its content.
      ``exit_on_error``: Boolean. Optional. If set to ``False``, any error encountered will only be logged as a warning. Default: ``True``.
    """
    logging.info(f"Deleting remote directory: {remote_dir}")
    try:
      for r_file in self.list_remote_dir(remote_dir=remote_dir):
        self.delete_remote_file(remote_file=r_file)
      if recursive:
        for r_dir in self.list_remote_dir(remote_dir=remote_dir, file_type="d"):
          self.delete_remote_dir(remote_dir=r_dir,
                                 recursive=recursive,
                                 exit_on_error=exit_on_error)
      self.client.rmd(remote_dir)
    except Exception as err:
      err_msg = f"[FtpTransferError] {err.__class__.__name__} {err}"
      if exit_on_error:
        logging.critical(err_msg)
        raise LaboroError(err_msg) from err
      logging.warning(err_msg)

  def _create_remote_dirs(self, remote_dir, exit_on_error=True):
    """
    Create the specified directory structure on the remote host.

    Arguments:
      ``remote_dir``: String. The remote directory full path to create.
    """
    hierarchy = [d for d in remote_dir.split("/") if len(d) > 0]
    logging.info(f"Creating remote dir: {remote_dir}")
    r_dir = "/"
    while len(hierarchy) > 0:
      r_dir = os.path.join(r_dir, hierarchy.pop(0))
      rmt_dir = os.path.dirname(r_dir)
      if f"{r_dir}" not in self.list_remote_dir(remote_dir=rmt_dir,
                                                file_type="d"):
        try:
          self.client.mkd(r_dir)
        except Exception as err:
          err_msg = f"[FtpCreateDirError] {err.__class__.__name__}: {err}"
          if exit_on_error:
            logging.critical(err_msg)
            raise LaboroError(err_msg) from err
          logging.warning(err_msg)



import logging
import socket
import ssl
from laboro_ftp.clients.errors import Error
from laboro_ftp.clients.ftp import FtpSimple, B_CRLF, CRLF, print_line


class Ftps(FtpSimple):
  """
  Ftp client with implicit TLS support (FTPs).
  """

  ssl_version = None
  ssl_verify = True
  context = None
  _prot_p = False

  def __init__(self, hostname, port=21,
               username='anonymous', password='anonymous',
               ssl_verify=True, ssl_version="TLSv1_2"):
    super().__init__(hostname=hostname, port=port,
                     username=username, password=password)
    self.ssl_verify = ssl_verify
    self.ssl_version = ssl.PROTOCOL_TLS_CLIENT
    if ssl_version is not None:
      self.ssl_version = getattr(ssl.TLSVersion, ssl_version)
    self.context = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH)
    if not self.ssl_verify:
      self.context.check_hostname = False
      self.context.verify_mode = ssl.CERT_NONE

  def connect(self):
    logging.debug("[+] Connection (Implicit TLS)")
    if self.hostname is not None:
      self.sock = socket.create_connection((self.hostname, self.port),
                                           self.timeout)
      self.af = self.sock.family
      if not self.ssl_verify:
        logging.warning("[!] Insecure connection: ssl_verify set to False")
      self.sock = self.context.wrap_socket(self.sock,
                                           server_hostname=self.hostname)
      self.file = self.sock.makefile('r', encoding=self.ENCODING)
      logging.debug(f"[WELCOME] {self._getresp()}")
      logging.debug("[=] connected")
      return True
    raise Error("Remote host not defined")

  def ntransfercmd(self, cmd, rest=None):
    conn, size = super().ntransfercmd(cmd, rest)
    if self._prot_p:
      conn = self.context.wrap_socket(conn,
                                      server_hostname=self.hostname,
                                      session=self.sock.session)
    return conn, size

  def storbinary(self, cmd, fp, blocksize=8192, callback=None, rest=None):
    self.prot_p()
    self.voidcmd('TYPE I')
    conn = self.transfercmd(cmd, rest)
    while 1:
      buf = fp.read(blocksize)
      if not buf:
        logging.debug("[STOR] EOF")
        break
      conn.sendall(buf)
      if callback:
        callback(buf)
    conn.unwrap()
    return self._voidresp()

  def storlines(self, cmd, fp, callback=None):
    self.prot_p()
    self.voidcmd('TYPE A')
    conn = self.transfercmd(cmd)
    while 1:
      buf = fp.readline(self.maxline + 1)
      if len(buf) > self.maxline:
        raise Error(f"got more than {self.maxline} bytes")
      if not buf:
        break
      if buf[-2:] != B_CRLF:
        if buf[-1] in B_CRLF:
          buf = buf[:-1]
          buf = buf + B_CRLF
      conn.sendall(buf)
      if callback:
        callback(buf)
    conn.unwrap()
    return self._voidresp()

  def retrbinary(self, cmd, callback, blocksize=8192, rest=None):
    self.prot_p()
    self.voidcmd('TYPE I')
    conn = self.transfercmd(cmd, rest)
    while 1:
      data = conn.recv(blocksize)
      if not data:
        break
      callback(data)
    conn.unwrap()
    return self._voidresp()

  def retrlines(self, cmd, callback=None):
    self.prot_p()
    if callback is None:
      callback = print_line
    resp = self.sendcmd('TYPE A')
    conn = self.transfercmd(cmd)
    with conn.makefile('r', encoding=self.ENCODING) as fp:
      while 1:
        line = fp.readline(self.maxline + 1)
        if len(line) > self.maxline:
          raise Error(f"got more than {self.maxline} bytes")
        logging.debug(f"[RETR] {repr(line)}")
        if not line:
          break
        if line[-2:] == CRLF:
          line = line[:-2]
        elif line[-1:] == '\n':
          line = line[:-1]
        callback(line)
    conn.unwrap()
    return self._voidresp()

  def close(self):
    logging.debug("[CLOSE]")
    self.file.close()
    try:
      self.sock.unwrap()
    except Exception:
      pass
    self.sock.close()

  def ccc(self):
    resp = self.voidcmd('CCC')
    self.sock = self.sock.unwrap()
    return resp

  def prot_p(self):
    self.voidcmd('PBSZ 0')
    resp = self.voidcmd('PROT P')
    self._prot_p = True
    return resp

  def prot_c(self):
    resp = self.voidcmd('PROT C')
    self._prot_p = False
    return resp



import logging
import socket
import ssl
from laboro_ftp.clients.errors import Error
from laboro_ftp.clients.ftps import Ftps


class Ftpes(Ftps):
  """
  Ftp client with explicit TLS support (FTPes).
  """

  ssl_version = None
  ssl_verify = True
  ssl_session = None
  context = None
  _prot_p = False

  def __init__(self, hostname, port=21,
               username='anonymous', password='anonymous',
               ssl_verify=True, ssl_version=None):
    super().__init__(hostname=hostname, port=port,
                     username=username, password=password)
    self.ssl_verify = ssl_verify
    self.ssl_version = ssl.PROTOCOL_TLS_CLIENT
    if ssl_version is not None:
      self.ssl_version = getattr(ssl.TLSVersion, ssl_version)
    self.context = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH)
    if not self.ssl_verify:
      self.context.check_hostname = False
      self.context.verify_mode = ssl.CERT_NONE

  def connect(self):
    logging.debug("[+] Connection (Explicit TLS)")
    if self.hostname is not None:
      self.sock = socket.create_connection((self.hostname, self.port),
                                           self.timeout)
      self.af = self.sock.family
      if not self.ssl_verify:
        logging.warning("[!] Insecure connection: ssl_verify set to False")
      self.file = self.sock.makefile('r', encoding=self.ENCODING)
      logging.debug(f"[WELCOME] {self._getresp()}")
      self.auth()
      logging.debug("[=] connected")
      return True
    raise Error("Remote host not defined")

  def auth(self):
    if isinstance(self.sock, ssl.SSLSocket):
      raise ValueError("Connection is already secured")
    if self.ssl_version >= ssl.PROTOCOL_TLS:
      resp = self.voidcmd('AUTH TLS')
    else:
      resp = self.voidcmd('AUTH SSL')
    self.sock = self.context.wrap_socket(self.sock,
                                         server_hostname=self.hostname)
    self.file = self.sock.makefile(mode='r', encoding=self.ENCODING)
    return resp
